lm.grp <- function(XtX, Xty, group, lambda, beta, mu,
                   eps.outer, eps.inner, max.it, optim.func){
  
  ## INITIALIZATION
  max.dL.min.old <- Inf
  p  <- ncol(XtX)
  pk <- tabulate(factor(group))
  active <- group.norm.rep(beta, pk) > 0
  iter.optim <- c()
  
  ## Check if optimality is reached at starting point
  dL.min <- dL.group.gaussian(beta,XtX,Xty,lambda,pk)
  n2.dL  <- group.norm(dL.min,pk)
  if (max(n2.dL) < eps.outer*p) {
    return(list(beta=beta,iter.optim=iter.optim,status="converged"))
  } else {
    ## UPDATE THE ACTIVE SET / ACTIVE GROUP
    new <- which.max(n2.dL)
    if (all(active[group == new])) {
      eps.optim  <- eps.outer
    } else {
      active[group == new] <- TRUE
      eps.optim  <- eps.inner
    }
  }

  while (1) {

    ## (1) OPTIMIZATION OVER A
    ## _____________________________________________________________
    res <- optim.func(as.matrix(beta[active]), XtX[active,active], Xty[active],
                      lambda[active],tabulate(factor(group[active])),eps.optim)
    iter.optim <- c(iter.optim,res$i)
    beta[active] <- res$xk
    ## dL.min only changes in the set of optimized parameters
    dL.min[active] <- dL.group.gaussian(beta[active],XtX[active,active],
                                        Xty[active],lambda[active],
                                        tabulate(factor(group[active])))
    
    ## (2) GROUP DELETION IF APPLICABLE
    ## _____________________________________________________________
    n2.beta <- group.norm(beta,pk)
    go.zero <- intersect(which(n2.beta < eps.optim),group[active])
    if (length(go.zero) > 0) {
      for (i in go.zero) {
        beta.tmp <- beta
        beta.tmp[group == i] <- 0
        n2.dL   <- group.norm(dL.group.gaussian(beta.tmp,XtX,Xty,lambda,pk),pk)
        if (n2.dL[i] < eps.outer*p) {
          active[group == i] <- FALSE
          beta[group == i] <- 0
        }
      }
    }

    ## (3) OPTIMALITY TESTING AND GROUP ACTIVATION IF APPLICABLE
    ## _____________________________________________________________
    max.dL.min <- max(dL.min)
    n2.dL  <- group.norm(dL.min,pk)
    if (length(iter.optim) == max.it) {
      return(list(beta=beta,iter.optim=iter.optim,status="max # of iterates reached",
                  max.n2dL=max(n2.dL)))
    }
    if (abs(max.dL.min - max.dL.min.old) < eps.outer) {
      return(list(beta=beta,iter.optim=iter.optim,status="no further improvement...",
                  max.n2dL=max.dL.min))
    }
    if (max(n2.dL) < eps.outer*p) {
      return(list(beta=beta,iter.optim=iter.optim,status="converged"))
    } else {
      ## UPDATE THE ACTIVE SET / ACTIVE GROUP
      max.dL.min.old <- max.dL.min
      new <- which.max(n2.dL)
      if (all(active[group == new])) {
        eps.optim <- eps.outer
      } else {
        active[group == new] <- TRUE
        eps.optim <- eps.inner
      }
    }
  }
}

L.group.gaussian <- function(beta,XtX,Xty,lambda,pk) {
  
  L <- .5*crossprod(beta,crossprod(XtX,beta))-crossprod(beta,Xty) +
    sum(lambda/rep.int(pk,pk) * group.norm.rep(beta,pk))

  return(L)
}

dL.group.gaussian <- function(beta,XtX,Xty,lambda,pk) {

  n2 <- group.norm.rep(beta,pk)
  zero <- n2 <= 0

  ## terms whose norm2 is different from zero
  dL <- rep.int(0,length(beta))

  nabla.f <- drop(crossprod(XtX,beta)) - Xty
  dL[!zero] <- nabla.f[!zero] + lambda[!zero] * beta[!zero] / n2[!zero]

  ## terms whose norm2 is zero
  n2.nabla.f <- group.norm.rep(nabla.f,pk)
  to.zero    <- n2.nabla.f[zero] <= lambda[zero]

  ## which ones can be null ?
  dL[zero][to.zero] <- 0
  dL[zero][!to.zero] <- nabla.f[zero][!to.zero] * (1 - lambda[zero][!to.zero]/n2.nabla.f[zero][!to.zero])

  return(dL)
}  

lm.coo <- function(XtX, Xty, group, lambda, beta, mu,
                   eps.outer, eps.inner, max.it, optim.func){

  ## INITIALIZATION
  p <- ncol(XtX)
  max.dL.min.old <- Inf
  pk <- tabulate(factor(group))
  active.pos <- group.norm.rep(pmax.int(0, beta),pk) > 0
  active.neg <- group.norm.rep(pmax.int(0,-beta),pk) > 0
  iter.optim <- c()

  ## Check if optimality is reached at the starting point
  dL.min <- dL.coop.gaussian(beta,XtX,Xty,lambda,pk) 
  n2.dL.pos  <- group.norm(pmax.int(0, dL.min),pk)
  n2.dL.neg  <- group.norm(pmax.int(0,-dL.min),pk)
  pmax.n2.dL <- pmax.int(n2.dL.pos,n2.dL.neg)
  if (max(pmax.n2.dL)<eps.outer*p) {
    return(list(beta=beta,iter.optim=iter.optim,status="converged"))
  } else {
    ## UPDATE THE ACTIVE SET / ACTIVE GROUP
    new <- which.max(pmax.n2.dL)
    if (all(active.neg[group == new]) | all(active.pos[group == new])) {
        eps.optim <- eps.outer
    } else {
        eps.optim <- eps.inner
    }
    if (n2.dL.pos[new] > n2.dL.neg[new]) {
      active.neg[group == new] <- TRUE
    } else {
      active.pos[group == new] <- TRUE
    }
  }
  
  while (1) {
    
    ## (1) OPTIMIZATION OVER A
    ## _____________________________________________________________
    lower.bound <- rep(-Inf,p)
    upper.bound <- rep(+Inf,p)
    lower.bound[active.pos & !active.neg] <- 0
    upper.bound[active.neg & !active.pos] <- 0
    active <- active.pos | active.neg
    res <- optim.func(as.matrix(beta[active]), XtX[active,active], Xty[active],
                      lambda[active],tabulate(factor(group[active])),eps.optim,
                      lower = lower.bound[active],upper=upper.bound[active])
    iter.optim <- c(iter.optim,res$i)
    beta[active] <- res$xk
    ## dL.min only changes in the set of optimized parameters
    dL.min[active] <- dL.coop.gaussian(beta[active],XtX[active,active],
                                       Xty[active],lambda[active],
                                       tabulate(factor(group[active])))
    
    ## (2) TEST GROUP DELETION IF APPLICABLE
    ## _____________________________________________________________
    n2.beta.pos <- group.norm(pmax.int(0,beta),pk)
    go.zero.pos <- intersect(which(n2.beta.pos < eps.optim),group[active.pos])
    if (length(go.zero.pos) > 0) {
      for (i in go.zero.pos) {
        beta.tmp <- beta
        beta.tmp[group == i & beta > 0] <- 0
        dL.min.tmp <- dL.coop.gaussian(beta.tmp,XtX,Xty,lambda,pk)
        n2.dL.neg <- group.norm(pmax.int(0,-dL.min.tmp),pk)
        if (n2.dL.neg[i] < eps.outer*p) {
          active.pos[group == i] <- FALSE
          beta[beta > 0 & group == i] <- 0
        }
      }
    }

    n2.beta.neg <- group.norm(pmax.int(0,-beta),pk)
    go.zero.neg <- intersect(which(n2.beta.neg < eps.optim),group[active.neg])
    if (length(go.zero.neg) > 0) {
      for (i in go.zero.neg) {
        beta.tmp <- beta
        beta.tmp[group == i & beta < 0] <- 0
        dL.min.tmp <- dL.coop.gaussian(beta.tmp,XtX,Xty,lambda,pk)
        n2.dL.pos <- group.norm(pmax.int(0, dL.min.tmp),pk)
        if (n2.dL.pos[i] < eps.outer*p) {
          active.neg[group == i] <- FALSE
          beta[beta < 0 & group == i] <- 0
        }
      }
    }
    
    ## (3) OPTIMALITY TESTING AND GROUP ACTIVATION IF APPLICABLE
    ## _____________________________________________________________
    ## Get the entry for which the constraint is the most violated
    ##    active <- active.pos | active.neg
    max.dL.min <- max(abs(dL.min))
    n2.dL.pos  <- group.norm(pmax.int(0, dL.min),pk)
    n2.dL.neg  <- group.norm(pmax.int(0,-dL.min),pk)
    pmax.n2.dL <- pmax.int(n2.dL.pos,n2.dL.neg)
    if (length(iter.optim) == max.it) {
      return(list(beta=beta,iter.optim=iter.optim,status="max # of iterates reached",
                  max.n2dL=max(pmax.n2.dL)))
    }
    if (abs(max.dL.min - max.dL.min.old) < eps.outer) {
      return(list(beta=beta,iter.optim=iter.optim,status="no further improvement...",
                  max.n2dL=max.dL.min))
    }
    if (max(pmax.n2.dL)<eps.outer*p) {
      return(list(beta=beta,iter.optim=iter.optim,status="converged"))
    } else {
      ## UPDATE THE ACTIVE SET / ACTIVE GROUP
      max.dL.min.old <- max.dL.min
      new <- which.max(pmax.n2.dL)
      if (all(active.neg[group == new]) | all(active.pos[group == new])) {
        eps.optim <- eps.outer
      } else {
        eps.optim <- eps.inner
      }
      if (n2.dL.pos[new] > n2.dL.neg[new]) {
        active.neg[group == new] <- TRUE
      } else {
        active.pos[group == new] <- TRUE
      }
    }
  }
}

L.coop.gaussian <- function(beta,XtX,Xty,lambda,pk) {

  ncoop <- group.norm.rep(pmax.int(0,beta),pk)+group.norm.rep(pmax.int(0,-beta),pk)
  L <- .5*crossprod(beta,crossprod(XtX,beta))-crossprod(beta,Xty) + sum(lambda/pk * ncoop)
  return(L)
}

dL.coop.gaussian <- function(beta,XtX,Xty,lambda,pk) {

  a.pos <- group.norm.rep(pmax.int(0, beta),pk)
  a.neg <- group.norm.rep(pmax.int(0,-beta),pk)
  
  ## manage cases where alpha+ > 0 or alpha- < 0
  dL <- drop(crossprod(XtX,beta)) - Xty
  neg <- beta < 0
  pos <- beta > 0
  dL[pos] <- dL[pos] + lambda[pos]*pmax.int(rep(0,sum(pos)), beta[pos]) / a.pos[pos]
  dL[neg] <- dL[neg] - lambda[neg]*pmax.int(rep(0,sum(neg)),-beta[neg]) / a.neg[neg]
  
  ## manage cases where beta = 0 and (alpha- = 0 and/or alpha+ = 0)
  n2.theta.pos <- group.norm.rep(pmax.int(0, dL),pk)
  n2.theta.neg <- group.norm.rep(pmax.int(0,-dL),pk)
  
  zero.pos  <- abs(beta) == 0 & dL >= 0 & a.neg == 0
  zero.neg  <- abs(beta) == 0 & dL <= 0 & a.pos == 0
  
  pos.to.zero  <- n2.theta.pos[zero.pos] <= lambda[zero.pos]
  neg.to.zero  <- n2.theta.neg[zero.neg] <= lambda[zero.neg]

  dL[zero.pos][!pos.to.zero] <- dL[zero.pos][!pos.to.zero] * (1-lambda[zero.pos][!pos.to.zero]/n2.theta.pos[zero.pos][!pos.to.zero])
  dL[zero.neg][!neg.to.zero] <- dL[zero.neg][!neg.to.zero] * (1-lambda[zero.neg][!neg.to.zero]/n2.theta.neg[zero.neg][!neg.to.zero])
  
  dL[zero.pos][pos.to.zero] <- 0
  dL[zero.neg][neg.to.zero] <- 0
  
  return(dL)
}

