lrm.grp <- function(X, y, group, lambda, beta, mu,
                    eps.outer, eps.inner, max.it, optim.func){
  
  ## INITIALIZATION
  p  <- ncol(X)
  pk <- tabulate(factor(group))
  active <- group.norm.rep(beta, pk) > 0
  iter.optim <- c()

  ## Check if optimality is reached at starting point
  dL.min <- dL.group.binomial(c(mu,beta),X,y,lambda,pk)
  if (max(abs(dL.min)) < eps.outer*p) {
    return(list(beta=beta,mu=mu,iter.optim=iter.optim,status="converged"))
  } else {
    ## UPDATE THE ACTIVE SET / ACTIVE GROUP
    if (is.null(mu)) {
      n2.dL <- group.norm(dL.min,pk)
    } else {
      n2.dL <- group.norm(dL.min[-1],pk)
    }
    new <- which.max(n2.dL)
    if (all(active[group == new])) {
      eps.optim  <- eps.outer
    } else {
      active[group == new] <- TRUE
      eps.optim  <- eps.inner
    }
  }

  while (1) {
    
    ## (1) OPTIMIZATION OVER A
    ## _____________________________________________________________
    res <- optim.func(as.matrix(c(mu,beta[active])), X[,active], y,
                      lambda[active], tabulate(factor(group[active])),eps.optim)
    iter.optim <- c(iter.optim,res$i)
    if (is.null(mu)) {
      beta[active] <- res$xk
      dL.min[active] <- dL.group.binomial(c(mu,beta[active]),X[,active],y,
                                          lambda[active],
                                          tabulate(factor(group[active])))
    } else {
      mu <- res$xk[1]
      beta[active] <- res$xk[-1]
      dL.min[c(TRUE,active)] <- dL.group.binomial(c(mu,beta[active]),X[,active],y,
                                                  lambda[active],
                                                  tabulate(factor(group[active])))
    }
    
    ## (2) TEST LAGRAGIAN VIOLATION AND GROUP DELETION IF APPLICABLE
    ## _____________________________________________________________
    go.zero <- intersect(which(group.norm(beta,pk) < eps.optim),group[active])
    if (length(go.zero) > 0) {
      for (i in go.zero) {
        beta.tmp <- beta
        beta.tmp[group == i] <- 0
        n2.dL  <- group.norm(dL.group.binomial(beta,X,y,lambda,pk),pk)
        if (n2.dL[i] < eps.optim) {
          active[group == i] <- FALSE
          beta[group == i] <- 0
        }
      }
    }

    ## (3) OPTIMALITY TESTING AND GROUP ACTIVATION IF APPLICABLE
    ## _____________________________________________________________
###    dL.min <- dL.group.binomial(c(mu,beta),X,y,lambda,pk)
    if (length(iter.optim) == max.it) {
      return(list(beta=beta,mu=mu,iter.optim=iter.optim,
                  status="max # of iterates reached",max.n2dL=max(n2.dL)))
    }    
    if (max(abs(dL.min)) < eps.outer*p) {
      return(list(beta=beta,mu=mu,iter.optim=iter.optim,status="converged"))
    } else {
      ## UPDATE THE ACTIVE SET / ACTIVE GROUP
      if (is.null(mu)) {
        n2.dL <- group.norm(dL.min,pk)
      } else {
        n2.dL  <- group.norm(dL.min[-1],pk)
      }
      new <- which.max(n2.dL)
      if (all(active[group == new])) {
        eps.optim <- eps.outer
      } else {
        active[group == new] <- TRUE
        eps.optim <- eps.inner
      }
    }
  }
}

L.group.binomial <- function(beta,X,y,lambda,pk) {

  ## Managing intercept
  if (length(beta) == length(lambda)) {
    eta <- X %*% beta
    n2 <- group.norm.rep(beta,pk) / rep.int(pk,pk)
  } else {
    eta <- cbind(1,X) %*% beta
    n2 <- group.norm.rep(beta[-1],pk) / rep.int(pk,pk)
  }
  
  L  <- - sum( y * eta - log(1 + exp(eta)) ) + sum(lambda *n2)

  return(L)
}

dL.group.binomial <- function(beta,X,y,lambda,pk) {

  ## Managing intercept
  if (length(beta) == length(lambda)) {
    eta <- X %*% beta
    nabla.f <- - t(X) %*% (y-1/(1+exp(-eta)))
    n2.nabla.f <- group.norm.rep(nabla.f,pk)
    n2 <- group.norm.rep(beta,pk)
  } else {
    eta <- cbind(1,X) %*% beta
    nabla.f <- - ( t(cbind(1,X)) %*% (y-1/(1+exp(-eta))) )
    n2.nabla.f <- c(abs(nabla.f[1]),group.norm.rep(nabla.f[-1],pk))
    lambda <- c(0,lambda)
    n2 <- c(abs(beta[1]),group.norm.rep(beta[-1],pk))
  }

  zero <- n2 == 0
  ## terms whose norm2 is different from zero
  dL <- rep(0,length(beta))
  dL[!zero] <- nabla.f[!zero] + lambda[!zero] * beta[!zero] / n2[!zero]
  
  ## terms whose norm2 is zero
  to.zero    <- n2.nabla.f[zero] <= lambda[zero]
  
  ## which ones can be null ?
  dL[zero][to.zero] <- 0
  dL[zero][!to.zero] <- nabla.f[zero][!to.zero] * (1 - lambda[zero][!to.zero]/n2.nabla.f[zero][!to.zero])

  return(dL)
}  

lrm.coo <- function(X, y, group, lambda, beta, mu,
                    eps.outer, eps.inner, max.it, optim.func){
  
  ## INITIALIZATION
  max.n2.dL.old <- Inf
  p  <- ncol(X)
  pk <- tabulate(factor(group))
  active.pos <- group.norm.rep(pmax.int(0, beta),pk) > 0
  active.neg <- group.norm.rep(pmax.int(0,-beta),pk) > 0
  iter.optim <- c()
  
  ## Check if optimality is reached at the starting point
  dL.min <- dL.coop.binomial(c(mu,beta),X,y,lambda,pk)
  if (max(abs(dL.min)) < eps.outer*p) {
    return(list(beta=beta,mu=mu,iter.optim=iter.optim,status="converged"))
  } else {
    ## UPDATE THE ACTIVE SET / ACTIVE GROUP
    if (is.null(mu)) {
      n2.dL.pos <- group.norm(pmax.int(0, dL.min),pk)
      n2.dL.neg <- group.norm(pmax.int(0,-dL.min),pk)
    } else {
      n2.dL.pos <- group.norm(pmax.int(0, dL.min[-1]),pk)
      n2.dL.neg <- group.norm(pmax.int(0,-dL.min[-1]),pk)
    }
    new <- which.max(pmax(n2.dL.pos,n2.dL.neg))
    if (all(active.neg[group == new]) | all(active.pos[group == new])) {
        eps.optim <- eps.outer
    } else {
        eps.optim <- eps.inner
    }
    if (n2.dL.pos[new] > n2.dL.neg[new]) {
      active.neg[group == new] <- TRUE
    } else {
      active.pos[group == new] <- TRUE
    }
  }
  
  while (1) {

    ## (1) OPTIMIZATION OVER A
    ## _____________________________________________________________
    lower.bound <- rep(-Inf,p)
    upper.bound <- rep(+Inf,p)
    lower.bound[active.pos & !active.neg] <- 0
    upper.bound[active.neg & !active.pos] <- 0
    active <- active.pos | active.neg
    res <- optim.func(as.matrix(c(mu,beta[active])), X[,active], y,
                      lambda[active],tabulate(factor(group[active])),eps.optim,
                      lower = lower.bound[active], upper = upper.bound[active])
    iter.optim <- c(iter.optim,res$i)
    if (is.null(mu)) {
      beta[active] <- res$xk
      dL.min[active] <- dL.coop.binomial(c(mu,beta[active]),X[,active],y,
                                         lambda[active],
                                         tabulate(factor(group[active])))
    } else {
      mu <- res$xk[1]
      beta[active] <- res$xk[-1]
      dL.min[c(TRUE,active)] <- dL.coop.binomial(c(mu,beta[active]),X[,active],y,
                                                 lambda[active],
                                                 tabulate(factor(group[active])))
    }
    
    ## (2) TEST LAGRAGIAN VIOLATION AND GROUP DELETION IF APPLICABLE
    ## _____________________________________________________________
    n2.beta.pos <- group.norm(pmax.int(0,beta),pk)
    go.zero.pos <- intersect(which(n2.beta.pos < eps.optim),group[active.pos])
    if (length(go.zero.pos) > 0) {
      for (i in go.zero.pos) {
        beta.tmp <- beta
        beta.tmp[group == i & beta > 0] <- 0
        dL.min.tmp <- dL.coop.binomial(beta.tmp,X,y,lambda,pk)
        n2.dL.neg <- group.norm(pmax.int(0,-dL.min.tmp),pk)
        if (n2.dL.neg[i] < eps.outer*p) {
          active.pos[group == i] <- FALSE
          beta[beta > 0 & group == i] <- 0
        }
      }
    }

    n2.beta.neg <- group.norm(pmax.int(0,-beta),pk)
    go.zero.neg <- intersect(which(n2.beta.neg < eps.optim),group[active.neg])
    if (length(go.zero.neg) > 0) {
      for (i in go.zero.neg) {
        beta.tmp <- beta
        beta.tmp[group == i & beta < 0] <- 0
        dL.min.tmp <- dL.coop.binomial(beta.tmp,X,y,lambda,pk)
        n2.dL.pos <- group.norm(pmax.int(0, dL.min.tmp),pk)
        if (n2.dL.pos[i] < eps.outer*p) {
          active.neg[group == i] <- FALSE
          beta[beta < 0 & group == i] <- 0
        }
      }
    }
    
    ## (3) OPTIMALITY TESTING AND GROUP ACTIVATION IF APPLICABLE
    ## _____________________________________________________________
    ## Get the entry for which the constraint is the most violated
###    dL.min <- dL.coop.binomial(c(mu,beta),X,y,lambda,pk)
    if (is.null(mu)) {
      n2.dL.pos <- group.norm(pmax.int(0, dL.min),pk)
      n2.dL.neg <- group.norm(pmax.int(0,-dL.min),pk)
    } else {
      n2.dL.pos <- group.norm(pmax.int(0, dL.min[-1]),pk)
      n2.dL.neg <- group.norm(pmax.int(0,-dL.min[-1]),pk)
    }
    pmax.n2.dL <- pmax.int(n2.dL.pos,n2.dL.neg)
    max.n2.dL  <- max(pmax.n2.dL)
    if (length(iter.optim) == max.it) {
      return(list(beta=beta,mu=mu,iter.optim=iter.optim,status="max # of iterates reached",
                  max.n2dL=max.n2.dL))
    }
    if (abs(max.n2.dL - max.n2.dL.old) < eps.outer) {
      return(list(beta=beta,mu=mu,iter.optim=iter.optim,status="no further improvement...",
                  max.n2dL=max.n2.dL))
    }
    
    if (max(abs(dL.min)) < eps.outer*p) {
      return(list(beta=beta,mu=mu,iter.optim=iter.optim,status="converged"))
    } else {
      max.n2.dL.old <- max.n2.dL
      ## UPDATE THE ACTIVE SET / ACTIVE GROUP
      new <- which.max(pmax.n2.dL)
      if (all(active.neg[group == new]) | all(active.pos[group == new])) {
        eps.optim <- eps.outer
      } else {
        eps.optim <- eps.inner
      }
      if (n2.dL.pos[new] > n2.dL.neg[new]) {
        active.neg[group == new] <- TRUE
      } else {
        active.pos[group == new] <- TRUE
      }
    }
  }
}

L.coop.binomial <- function(beta,X,y,lambda,pk) {

  ## Managing intercept
  if (length(beta) == length(lambda)) {
    eta <- X %*% beta
    a.pos <- group.norm.rep(pmax.int(0, beta),pk)/rep.int(pk,pk)
    a.neg <- group.norm.rep(pmax.int(0,-beta),pk)/rep.int(pk,pk)
  } else {
    eta <- cbind(1,X) %*% beta
    a.pos <- group.norm.rep(pmax.int(0, beta[-1]),pk)/rep.int(pk,pk)
    a.neg <- group.norm.rep(pmax.int(0,-beta[-1]),pk)/rep.int(pk,pk)
  }
  
  L <- - sum( y * eta - log(1 + exp(eta)) ) + sum(lambda *(a.pos + a.neg))

  return(L)
}

dL.coop.binomial <- function(beta,X,y,lambda,pk) {

  ## Managing intercept
  if (length(beta) == length(lambda)) {
    inter <- FALSE
    eta <- X %*% beta
    nabla.f <- - t(X) %*% (y-1/(1+exp(-eta)))
    a.pos <- group.norm.rep(pmax.int(0, beta),pk)
    a.neg <- group.norm.rep(pmax.int(0,-beta),pk)
    n2.theta.pos <- group.norm.rep(pmax.int(0, nabla.f),pk)
    n2.theta.neg <- group.norm.rep(pmax.int(0,-nabla.f),pk)
  } else {
    inter <- TRUE
    eta <- cbind(1,X) %*% beta
    nabla.f <- - ( t(cbind(1,X)) %*% (y-1/(1+exp(-eta))) )
    lambda <- c(0,lambda)
    a.pos <- c(max(0, beta[1]),group.norm.rep(pmax.int(0, beta[-1]),pk))
    a.neg <- c(max(0,-beta[1]),group.norm.rep(pmax.int(0,-beta[-1]),pk))
    n2.theta.pos <- c(max(0, nabla.f[1]),group.norm.rep(pmax.int(0, nabla.f[-1]),pk))
    n2.theta.neg <- c(max(0,-nabla.f[1]),group.norm.rep(pmax.int(0,-nabla.f[-1]),pk))
  }
    
  ## manage cases where alpha+ > 0 or alpha- < 0
  dL <- nabla.f
  neg <- beta < 0
  pos <- beta > 0
  dL[pos] <- nabla.f[pos] + lambda[pos]*pmax.int(rep(0,sum(pos)), beta[pos]) / a.pos[pos]
  dL[neg] <- nabla.f[neg] - lambda[neg]*pmax.int(rep(0,sum(neg)),-beta[neg]) / a.neg[neg]
  
  ## manage cases where beta = 0 and (alpha- = 0 and/or alpha+ = 0)
  zero.pos  <- abs(beta) == 0 & nabla.f > 0 & a.neg == 0
  zero.neg  <- abs(beta) == 0 & nabla.f < 0 & a.pos == 0
  
  pos.to.zero  <- n2.theta.pos[zero.pos] <= lambda[zero.pos]
  neg.to.zero  <- n2.theta.neg[zero.neg] <= lambda[zero.neg]

  dL[zero.pos][!pos.to.zero] <- nabla.f[zero.pos][!pos.to.zero] * (1-lambda[zero.pos][!pos.to.zero]/n2.theta.pos[zero.pos][!pos.to.zero])
  dL[zero.neg][!neg.to.zero] <- nabla.f[zero.neg][!neg.to.zero] * (1-lambda[zero.neg][!neg.to.zero]/n2.theta.neg[zero.neg][!neg.to.zero])
  
  dL[zero.pos][pos.to.zero] <- 0
  dL[zero.neg][neg.to.zero] <- 0
  
  return(dL)
}
